def set_direction_reg(self, direction):
    pass

def set_current(self, run_current, hold_current_multiplier = 0.5, hold_current_delay = 10, pdn_disable = True):
    pass

def set_interpolation(self, en):
    pass

def set_spreadcycle(self,en_spread):
    pass

def set_microstepping_resolution(self, msres):
    pass

def set_internal_rsense(self,en):
    pass
