from enum import Enum

class MovementAbsRel(Enum):
    """movement absolute or relative"""
    ABSOLUTE = 0
    RELATIVE = 1

def set_max_speed(self, speed):
    pass

def set_acceleration(self, acceleration):
    pass

def run_to_position_revolutions(self, revolutions, movement_abs_rel = None):
    pass

def set_movement_abs_rel(self, movement_abs_rel):
    pass
